from distutils.core import setup

setup(
    name='pyqtlet2',
    packages=[
        'pyqtlet2',
        'pyqtlet2.web',
        'pyqtlet2.web.modules.leaflet_134',
        'pyqtlet2.web.modules.leaflet_134.images',
        'pyqtlet2.web.modules.leaflet_draw_104',
        'pyqtlet2.web.modules.leaflet_draw_104.dist',
        'pyqtlet2.web.modules.leaflet_draw_104.dist.images',
        'pyqtlet2.leaflet',
        'pyqtlet2.leaflet.control',
        'pyqtlet2.leaflet.core',
        'pyqtlet2.leaflet.layer',
        'pyqtlet2.leaflet.layer.marker',
        'pyqtlet2.leaflet.layer.tile',
        'pyqtlet2.leaflet.layer.vector',
        'pyqtlet2.leaflet.map',
    ],
    package_data={
        'pyqtlet2.web': ['*'], 'pyqtlet2.web.modules': ['*'],
        'pyqtlet2.web.modules.leaflet_134': ['*'],
        'pyqtlet2.web.modules.leaflet_134.images': ['*'],
        'pyqtlet2.web.modules.leaflet_draw_104': ['*'],
        'pyqtlet2.web.modules.leaflet_draw_104.dist': ['*'],
        'pyqtlet2.web.modules.leaflet_draw_104.dist.images': ['*'],
    },
    version='0.3.4',
    description='Bringing leaflet maps to PyQt',
    author='Leon Friedmann',
    author_email='leon.friedmann@tum.de',
    url='https://github.com/JaWeilBaum/pyqtlet2',
    keywords=['leaflet', 'pyqt', 'maps', 'python', 'python3'],
    classifiers=[],
)
